﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
// type definitions for ease of use
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// An operator that gathers information on tree fragments that get passed from one generation to the next via genetic operators
  /// (Tracking of genetic variance and heredity)
  /// </summary>
  /// [Item("SymbolicExpressionTreeGenealogyAnalyzer", "An operator that tracks tree lengths of Symbolic Expression Trees")]
  [StorableClass]
  public sealed class SymbolicExpressionTreeFragmentsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string GlobalTraceMapParameterName = "GlobalTraceMap";
    private const string GlobalCloneMapParameterName = "GlobalCloneMap";
    private const string GenerationsParameterName = "Generations";
    private const string FragmentStatisticsParameterName = "FragmentStatistics";

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public LookupParameter<DataTable> FragmentStatisticsParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentStatisticsParameterName]; }

    }
    #endregion

    #region Parameters
    public bool EnabledByDefault {
      get { return true; }
    }
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public DataTable FragmentStatistics {
      get { return FragmentStatisticsParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreeFragmentsAnalyzer(bool deserializing)
      : base() {
    }
    private SymbolicExpressionTreeFragmentsAnalyzer(SymbolicExpressionTreeFragmentsAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreeFragmentsAnalyzer(this, cloner);
    }
    public SymbolicExpressionTreeFragmentsAnalyzer()
      : base() {
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName, "A global cache containing the whole genealogy."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName, "A global map keeping track of trees and their clones (made during selection)."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentStatisticsParameterName, "The data table to store the fragment statistics."));
      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }
    #region After deserialization code
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // check if all the parameters are present and accounted for 
      if (!Parameters.ContainsKey(UpdateIntervalParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      }
      if (Parameters.ContainsKey(UpdateCounterParameterName)) return;
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      UpdateCounterParameter.Hidden = true;
    }
    #endregion
    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    public override IOperation Apply() {
      if (Generations.Value == 0) return base.Apply();
      UpdateCounter.Value++;
      if (UpdateCounter.Value == UpdateInterval.Value) {
        ResultCollection results = ResultsParameter.ActualValue;
        if (FragmentStatistics == null) {
          FragmentStatisticsParameter.ActualValue = new DataTable("Fragment Statistics", "Statistical measurements of fragments aggregated over the whole population");
          FragmentStatistics.VisualProperties.YAxisTitle = "Fragment length/Similarities";
          results.Add(new Result("Fragment Statistics", FragmentStatistics));
        }

        int numberOfValues = FragmentStatistics.Rows.Select(r => r.Values.Count).DefaultIfEmpty().First();

        UpdateCounter.Value = 0; // reset counter
        var gScope = ExecutionContext.Scope;
        while (gScope.Parent != null) gScope = gScope.Parent;

        #region Fragment Statistics
        // write these to file for now, later we will use a DataTable and display the results inside HeuristicLab
        var path = Environment.GetFolderPath(Environment.SpecialFolder.Personal);
        //using (var file = System.IO.File.AppendText(path + @"\tracking.csv")) {
        var fragments = new List<ISymbolicExpressionTreeNode>();
        var parents = new List<ISymbolicExpressionTree>();
        double best = gScope.SubScopes.Select(x => x.Variables["Quality"].Value as DoubleValue).Max(x => x.Value);
        var trees = (from s in gScope.SubScopes select s.Variables.First().Value as ISymbolicExpressionTree).ToList();
        foreach (var tree in trees.Where(x => GlobalTraceMap.ContainsKey(x))) {
          var parent0 = (ISymbolicExpressionTree)GlobalTraceMap[tree][0];
          var fragment = SymbolicExpressionTreeMatching.GetFragmentDiff(parent0, tree);
          if (fragment != null) {
            parents.Add(parent0);
            fragments.Add(fragment);
            if (GlobalTraceMap.ContainsKey(parent0)) {
              var p = (ISymbolicExpressionTree)GlobalTraceMap[parent0][0];
              fragment = SymbolicExpressionTreeMatching.GetFragmentDiff(p, parent0);
              if (fragment != null) {
                parents.Add(p);
                fragments.Add(fragment);
              }
            }
          }
        }

        // write values to file
        double a1 = fragments.Average(x => x.GetLength());
        double a2 = trees.Average(x => x.Length);
        double a3 = parents.Average(x => x.Length);
        double s1 = CalculateSimilarity(fragments, (int)SymbolicExpressionTreeMatching.SimilarityLevel.Exact);
        double s2 = CalculateSimilarity(fragments, (int)SymbolicExpressionTreeMatching.SimilarityLevel.High);
        double s3 = CalculateSimilarity(fragments, (int)SymbolicExpressionTreeMatching.SimilarityLevel.Relaxed);

        #region Table data
        // fragments
        if (!FragmentStatistics.Rows.ContainsKey("Average fragment length")) {
          DataRow row = new DataRow("Average fragment length", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Average fragment length"].Values.Add(a1);
        // child trees
        if (!FragmentStatistics.Rows.ContainsKey("Average child length")) {
          DataRow row = new DataRow("Average child length", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Average child length"].Values.Add(a2);
        // parents
        if (!FragmentStatistics.Rows.ContainsKey("Average parent length")) {
          DataRow row = new DataRow("Average parent length", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Average parent length"].Values.Add(a3);
        // exact similarity
        if (!FragmentStatistics.Rows.ContainsKey("Similarity (exact)")) {
          DataRow row = new DataRow("Similarity (exact)", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Similarity (exact)"].Values.Add(s1);
        // exact similarity
        if (!FragmentStatistics.Rows.ContainsKey("Similarity (high)")) {
          DataRow row = new DataRow("Similarity (high)", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Similarity (high)"].Values.Add(s2);
        // exact similarity
        if (!FragmentStatistics.Rows.ContainsKey("Similarity (relaxed)")) {
          DataRow row = new DataRow("Similarity (relaxed)", "");
          row.VisualProperties.StartIndexZero = true;
          FragmentStatistics.Rows.Add(row);
        }
        FragmentStatistics.Rows["Similarity (relaxed)"].Values.Add(s3);

        //file.WriteLine(best * 10.0 + " " + a1 + " " + a2 + " " + a3 + " " + s1 + " " + s2 + " " + s3);
        //}
        #endregion
        #endregion
        // clear the global maps to save memory
        GlobalCloneMap.Clear();
        GlobalTraceMap.Clear();
      }
      return base.Apply();
    }

    #region Similarity computations
    #region Obsolete code
    private void CalculateDiversityIndicators(IEnumerable<ISymbolicExpressionTreeNode> fragments, out double a, out double b, out double c) {
      //var frags = fragments.Where(x => x.GetLength() > 1).ToList();
      var frags = fragments.ToList();
      var sieve = new bool[frags.Count];
      int max = 0;
      for (int i = 0; i != frags.Count - 1; ++i) {
        if (sieve[i]) continue;
        int count = 0;
        for (int j = i + 1; j != frags.Count; ++j) {
          if (sieve[j]) continue;
          if (frags[i].IsSimilarTo(frags[j], (int)SymbolicExpressionTreeMatching.SimilarityLevel.Exact)) {
            ++count;
            sieve[i] = true;
            sieve[j] = true;
          }
        }
        if (max < count) max = count;
      }
      a = (double)sieve.Count(x => x) / sieve.Length;
      sieve = new bool[frags.Count];
      max = 0;
      for (int i = 0; i != frags.Count - 1; ++i) {
        if (sieve[i]) continue;
        int count = 0;
        for (int j = i + 1; j != frags.Count; ++j) {
          if (sieve[j]) continue;
          if (frags[i].IsSimilarTo(frags[j], (int)SymbolicExpressionTreeMatching.SimilarityLevel.High)) {
            ++count;
            sieve[i] = true;
            sieve[j] = true;
          }
        }
        if (max < count) max = count;
      }
      b = (double)sieve.Count(x => x) / sieve.Length;
      sieve = new bool[frags.Count];
      max = 0;
      for (int i = 0; i != frags.Count - 1; ++i) {
        if (sieve[i]) continue;
        int count = 0;
        for (int j = i + 1; j != frags.Count; ++j) {
          if (sieve[j]) continue;
          if (frags[i].IsSimilarTo(frags[j], (int)SymbolicExpressionTreeMatching.SimilarityLevel.Relaxed)) {
            ++count;
            sieve[i] = true;
            sieve[j] = true;
          }
        }
        if (max < count) max = count;
      }
      c = (double)sieve.Count(x => x) / max;
    }
    #endregion

    private double CalculateSimilarity(List<ISymbolicExpressionTreeNode> fragments, int mode) {
      bool[] visited = new bool[fragments.Count];
      int count = 0;
      for (int i = 0; i != fragments.Count - 1; ++i) {
        if (visited[i]) continue;
        for (int j = i + 1; j != fragments.Count; ++j) {
          if (visited[j]) continue;
          if (fragments[i].IsSimilarTo(fragments[j], mode)) visited[j] = true;
        }
        ++count;
      }
      return (double)fragments.Count / count;
    }
    #endregion
  } //SymbolicExpressionTreeFragmentsAnalyzer
}
